%  Copyright (C) 2002-2004 David Roundy
%
%  This program is free software; you can redistribute it and/or modify
%  it under the terms of the GNU General Public License as published by
%  the Free Software Foundation; either version 2, or (at your option)
%  any later version.
%
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License for more details.
%
%  You should have received a copy of the GNU General Public License
%  along with this program; see the file COPYING.  If not, write to
%  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
%  Boston, MA 02110-1301, USA.

\darcsCommand{help}
\begin{code}
module Darcs.Commands.Help (
 help_cmd,
 command_control_list, environmentHelp,          -- these are for preproc.hs
 print_version,
 list_available_commands ) where

import Darcs.Arguments ( DarcsFlag(..), environmentHelpSendmail )
import Darcs.Commands (
 CommandArgs(..), CommandControl(..), DarcsCommand(..),
 disambiguate_commands, extract_commands, get_command_help, nodefaults, usage )
import Darcs.External ( viewDoc )
import Darcs.Lock ( environmentHelpTmpdir, environmentHelpKeepTmpdir )
import Darcs.Patch.Match ( helpOnMatchers )
import Darcs.Repository.Prefs ( binaries_file_help, environmentHelpHome )
import Darcs.Utils ( withCurrentDirectory, environmentHelpEditor, environmentHelpPager )
import Data.Char ( isAlphaNum, toLower )
import Data.List ( groupBy )
import English ( andClauses )
import Printer ( text )
import Ssh ( environmentHelpSsh, environmentHelpScp, environmentHelpSshPort )
import System.Exit ( ExitCode(..), exitWith )
import ThisVersion ( darcs_version )
import URL (environmentHelpProxy, environmentHelpProxyPassword)
import Workaround ( getCurrentDirectory )
import qualified Darcs.TheCommands as TheCommands

help_description :: String
help_description = "Display help about darcs and darcs commands."

help_help :: String
help_help =
 "Without arguments, `darcs help' prints a categorized list of darcs\n" ++
 "commands and a short description of each one.  With an extra argument,\n" ++
 "`darcs help foo' prints detailed help about the darcs command foo.\n"

help :: DarcsCommand
help = DarcsCommand {command_name = "help",
                     command_help = help_help,
                     command_description = help_description,
                     command_extra_args = -1,
                     command_extra_arg_help = ["[<DARCS_COMMAND> [DARCS_SUBCOMMAND]]  "],
                     command_command = \ x y -> help_cmd x y >> exitWith ExitSuccess,
                     command_prereq = \_ -> return $ Right (),
                     command_get_arg_possibilities = return [],
                     command_argdefaults = nodefaults,
                     command_advanced_options = [],
                     command_basic_options = []}

help_cmd :: [DarcsFlag] -> [String] -> IO ()
help_cmd _ ["manpage"] = putStr $ unlines manpageLines
help_cmd _ ["patterns"] = viewDoc $ text $ helpOnMatchers
help_cmd _ ["environment"] = viewDoc $ text $ helpOnEnvironment
help_cmd _ [] = viewDoc $ text $ usage command_control_list

help_cmd _ (cmd:args) =
    let disambiguated = disambiguate_commands command_control_list cmd args
    in case disambiguated of
         Left err -> fail err
         Right (cmds,_) ->
             let msg = case cmds of
                         CommandOnly c       -> get_command_help Nothing  c
                         SuperCommandOnly c  -> get_command_help Nothing  c
                         SuperCommandSub c s -> get_command_help (Just c) s
             in viewDoc $ text msg

list_available_commands :: IO ()
list_available_commands =
    do here <- getCurrentDirectory
       is_valid <- sequence $ map
                   (\c-> withCurrentDirectory here $ (command_prereq c) [])
                   (extract_commands command_control_list)
       putStr $ unlines $ map (command_name . fst) $
                filter (isRight.snd) $
                zip (extract_commands command_control_list) is_valid
       putStrLn "--help"
       putStrLn "--version"
       putStrLn "--exact-version"
       putStrLn "--overview"
    where isRight (Right _) = True
          isRight _ = False

print_version :: IO () 
print_version = putStrLn $ "darcs version " ++ darcs_version

-- avoiding a module import cycle between Help and TheCommands
command_control_list :: [CommandControl] 
command_control_list =
  Command_data help : TheCommands.command_control_list

-- | Help on each environment variable in which Darcs is interested.
environmentHelp :: [([String], [String])]
environmentHelp = [
 -- General-purpose
 environmentHelpHome,
 environmentHelpEditor,
 environmentHelpPager,
 environmentHelpTmpdir,
 environmentHelpKeepTmpdir,
 environmentHelpSendmail,
 -- Remote Repositories
 environmentHelpSsh,
 environmentHelpScp,
 environmentHelpSshPort,
 environmentHelpProxy,
 environmentHelpProxyPassword]

-- | The rendered form of the data in 'environment_help'.
helpOnEnvironment :: String
helpOnEnvironment =
    "Environment Variables\n" ++
    "=====================\n\n" ++
    unlines [andClauses ks ++ ":\n" ++
                     (unlines $ map ("  " ++) ds)
                     | (ks, ds) <- environmentHelp]

-- | This module is responsible for emitting a darcs "man-page", a
-- reference document used widely on Unix-like systems.  Manpages are
-- primarily used as a quick reference, or "memory jogger", so the
-- output should be terser than the user manual.
--
-- Before modifying the output, please be sure to read the man(7) and
-- man-pages(7) manpages, as these respectively describe the relevant
-- syntax and conventions.

-- | The lines of the manpage to be printed.
manpageLines :: [String]
manpageLines = [
 ".TH DARCS 1 \"" ++ darcs_version ++ "\"",
 ".SH NAME",
 "darcs \\- an advanced revision control system",
 ".SH SYNOPSIS",
 ".B darcs", ".I command", ".RI < arguments |[ options ]>...",
 "",
 "Where the", ".I commands", "and their respective", ".I arguments", "are",
 "",
 unlines synopsis,
 ".SH DESCRIPTION",
 -- FIXME: this is copy-and-pasted from darcs.cabal, so
 -- it'll get out of date as people forget to maintain
 -- both in sync.
 "Darcs is a free, open source revision control",
 "system. It is:",
 ".TP 3", "\\(bu",
 "Distributed: Every user has access to the full",
 "command set, removing boundaries between server and",
 "client or committer and non-committers.",
 ".TP", "\\(bu",
 "Interactive: Darcs is easy to learn and efficient to",
 "use because it asks you questions in response to",
 "simple commands, giving you choices in your work",
 "flow. You can choose to record one change in a file,",
 "while ignoring another. As you update from upstream,",
 "you can review each patch name, even the full `diff'",
 "for interesting patches.",
 ".TP", "\\(bu",
 "Smart: Originally developed by physicist David",
 "Roundy, darcs is based on a unique algebra of",
 "patches.",
 "This smartness lets you respond to changing demands",
 "in ways that would otherwise not be possible. Learn",
 "more about spontaneous branches with darcs.",
 ".SH OPTIONS",
 "Different options are accepted by different Darcs commands.",
 "Each command's most important options are listed in the",
 ".B COMMANDS",
 "section.  For a full list of all options accepted by",
 "a particular command, run `darcs", ".I command", "--help'.",
 ".SS " ++ helpOnMatchers, -- FIXME: this is a kludge.
 ".SH COMMANDS",
 unlines commands,
 unlines environment,
 ".SH FILES",
 ".SS \"_darcs/prefs/binaries\"",
 unlines binaries_file_help,
 ".SH BUGS",
 "At http://bugs.darcs.net/ you can find a list of known",
 "bugs in Darcs.  Unknown bugs can be reported at that",
 "site (after creating an account) or by emailing the",
 "report to bugs@darcs.net.",
 -- ".SH EXAMPLE",
 -- FIXME:
 -- new project: init, rec -la;
 -- track upstream project: get, pull -a;
 -- contribute to project: add, rec, push/send.
 ".SH SEE ALSO",
 "A user manual is included with Darcs, in PDF and HTML",
 "form.  It can also be found at http://darcs.net/manual/."]
    where
      -- | A synopsis line for each command.  Uses 'foldl' because it is
      -- necessary to avoid blank lines from Hidden_commands, as groff
      -- translates them into annoying vertical padding (unlike TeX).
      synopsis :: [String]
      synopsis = foldl iter [] command_control_list
          where iter :: [String] -> CommandControl -> [String]
                iter acc (Group_name _) = acc
                iter acc (Hidden_command _) = acc
                iter acc (Command_data c@SuperCommand {}) =
                    acc ++ concatMap
                            (render (command_name c ++ " "))
                            (extract_commands (command_sub_commands c))
                iter acc (Command_data c) = acc ++ render "" c
                render :: String -> DarcsCommand -> [String]
                render prefix c =
                    [".B darcs " ++ prefix ++ command_name c] ++
                    (map mangle_args $ command_extra_arg_help c) ++
                    -- In the output, we want each command to be on its own
                    -- line, but we don't want blank lines between them.
                    -- AFAICT this can only be achieved with the .br
                    -- directive, which is probably a GNUism.
                    [".br"]

      -- | As 'synopsis', but make each group a subsection (.SS), and
      -- include the help text for each command.
      commands :: [String]
      commands = foldl iter [] command_control_list
          where iter :: [String] -> CommandControl -> [String]
                iter acc (Group_name x) = acc ++ [".SS \"" ++ x ++ "\""]
                iter acc (Hidden_command _) = acc
                iter acc (Command_data c@SuperCommand {}) =
                    acc ++ concatMap
                            (render (command_name c ++ " "))
                            (extract_commands (command_sub_commands c))
                iter acc (Command_data c) = acc ++ render "" c
                render :: String -> DarcsCommand -> [String]
                render prefix c =
                    [".B darcs " ++ prefix ++ command_name c] ++
                    (map mangle_args $ command_extra_arg_help c) ++
                    [".RS 4", command_help c, ".RE"]

      -- | Now I'm showing off: mangle the extra arguments of Darcs commands
      -- so as to use the ideal format for manpages, italic words and roman
      -- punctuation.
      mangle_args :: String -> String
      mangle_args s =
          ".RI " ++ (unwords $ map show (groupBy cmp $ map toLower $ gank s))
              where cmp x y = not $ xor (isAlphaNum x) (isAlphaNum y)
                    xor x y = (x && not y) || (y && not x)
                    gank (' ':'o':'r':' ':xs) = '|' : gank xs
                    gank (x:xs) = x : gank xs
                    gank [] = []

      environment :: [String]
      environment = ".SH ENVIRONMENT" : concat
                    [(".SS \"" ++ andClauses ks ++ "\"") : ds
                     | (ks, ds) <- environmentHelp]

\end{code}
